;; -*- lexical-binding: t -*-


(defun stream-substream-before (stream rest)
  "Return a stream of the elements of STREAM before REST.

REST must either be `eq' to STREAM or to one of the subsequent
calls of `stream-rest' on STREAM."
  (stream-make
   (if (eq stream rest)
       nil
     (cons (stream-first stream)
           (stream-substream-before (stream-rest stream) rest)))))

(defun stream-divide-with-get-rest-fun (stream get-rest-fun)
  (let ((rest (funcall get-rest-fun stream)))
    (list (stream-substream-before stream rest) rest)))

(defun stream-divide (stream pred)
  "Divide STREAM between the first ELT, NEXT element pair for that PRED returns nil.
The return value is a list of two streams.  If no such pair of
elements exists, return (STREAM E) where E is an empty stream.
Example:

  (mapcar #'seq-into-sequence
          (stream-divide
           (stream (list 1 2 3 5 6 7 9 10 11 23))
           (lambda (this next) (< (- next this) 2))))
==> ((1 2 3)
     (5 6 7 9 10 11 23))"
  (stream-divide-with-get-rest-fun stream (stream-divide--get-rest-fun pred)))

(defun stream-divide--get-rest-fun (pred)
  (lambda (s)
    (unless (stream-empty-p s)
      (while (let ((this (stream-pop s)))
               (unless (stream-empty-p s)
                 (funcall pred this (stream-first s))))))
    s))


(defun stream-partition (stream pred)
  "Partition STREAM into bunches where PRED returns non-nil for subsequent elements.
The return value is a stream of streams (so this function is a
reverse of `stream-concatenate').

Often, but not necessarily, PRED is an equivalence predicate.
Example:

  (seq-into-sequence
   (seq-map #'seq-into-sequence
            (stream-partition
             (stream (list 1 2 3 5 6 7 9 10 15 23))
                (lambda (x y) (< (- y x) 2)))))
   ==> ((1 2 3)
        (5 6 7)
        (9 10)
        (15)
        (23))"
  (stream-partition-with-get-rest-fun stream (stream-divide--get-rest-fun pred)))

(defun stream-partition-with-get-rest-fun (stream get-rest-fun)
  (stream-make
   (if (stream-empty-p stream) nil
     (let ((divided (stream-divide-with-get-rest-fun stream get-rest-fun)))
       (cons (car divided)
             (stream-partition-with-get-rest-fun (cadr divided) get-rest-fun))))))
